from typing import List, Union, Optional
from uuid import UUID
from pyarrow import csv, Table

from datalogue.clients.http import _HttpClient, HttpMethod, MultiPartData, MultiPartFile
from datalogue.models.datastore import Datastore, _datastore_from_payload, FileFormat, Cell, _cell_from_payload

from datalogue.errors import DtlError
from datalogue.utils import _parse_list, ResponseStream


class _DatastoreClient:
    """
    Client to interact with the Datastores objects
    """

    def __init__(self, http_client: _HttpClient):
        self.http_client = http_client
        self.service_uri = "/scout"

    def create(self, datastore: Datastore) -> Union[DtlError, Datastore]:
        """
        Creates a data source

        :param datastore: data source to be created
        :return: string with error message if failed, the data source otherwise
        """
        res = self.http_client.make_authed_request(
            self.service_uri + "/datastores", HttpMethod.POST, datastore._as_payload())

        if isinstance(res, DtlError):
            return res

        return _datastore_from_payload(res)

    def update(self, datastore: Datastore) -> Union[DtlError, Datastore]:
        """
        Updates the backend with the new status of the existing datastore

        :param datastore: to be persisted
        :return:
        """

        res = self.http_client.make_authed_request(
            self.service_uri + "/datastores/" + str(datastore.id),
            HttpMethod.PUT,
            datastore._as_payload()
        )

        if isinstance(res, DtlError):
            return res

        return _datastore_from_payload(res)

    def list(self, page: int = 1, item_per_page: int = 25) -> Union[DtlError, List[Datastore]]:
        """
        List all the datastores that are saved

        TODO pagination

        :param page: page to be retrieved (ignored)
        :param item_per_page: number of items to be put in a page (ignored)
        :return: Returns a List of all the available datastores or an error message as a string
        """

        res = self.http_client.make_authed_request(self.service_uri + "/datastores", HttpMethod.GET)

        if isinstance(res, DtlError):
            return res

        return _parse_list(_datastore_from_payload)(res)

    def get_schema(self, datastore_id: UUID) -> Union[DtlError, List[List[Cell]]]:
        """
        From the provided id, get the corresponding Datastore schema

        :param datastore_id: id of the datastore to be retrieved
        :return:
        """
        res = self.http_client.make_authed_request(self.service_uri + "/datastores/" + str(datastore_id) + "/schema-samples", HttpMethod.GET)

        if isinstance(res, DtlError):
            return res

        return _parse_list(_parse_list(_cell_from_payload))(res['schema'])

    def get(self, datastore_id: UUID) -> Union[DtlError, Datastore]:
        """
        From the provided id, get the corresponding Datastore

        :param datastore_id: id of the datastore to be retrieved
        :return:
        """

        res = self.http_client.make_authed_request(self.service_uri + "/datastores/" + str(datastore_id), HttpMethod.GET)

        if isinstance(res, DtlError):
            return res

        return _datastore_from_payload(res)

    def delete(self, datastore_id: UUID) -> Union[DtlError, bool]:
        """
        Deletes the given Datastore

        :param datastore_id: id of the datastore to be deleted
        :return: true if successful, false otherwise
        """

        res = self.http_client.make_authed_request(
            self.service_uri + "/datastores/" + str(datastore_id), HttpMethod.DELETE)

        if isinstance(res, DtlError):
            return res
        else:
            return True

    def download(self, datastore_id: UUID, file_format: FileFormat, path: str, limit: Optional[int] = None) -> Union[DtlError, str]:
        """
        Downloads the resource in a file

        :param datastore_id: id of the resource to download
        :param file_format: File format of the downloaded resource
        :param path: location to save the file
        :param limit: limit the number of elements to be downloaded
        :return: An error or the path to the downloaded file
        """

        params = {
            "fileFormat": file_format.value,
        }

        if limit is not None:
            if not isinstance(limit, int) or limit < 0:
                return DtlError("limit should be a positive integer")

            params["limit"] = limit

        res = self.http_client.make_authed_request(
            self.service_uri + "/datastores/" + str(datastore_id) + "/download",
            HttpMethod.GET, params=params, stream=True
        )

        if isinstance(res, DtlError):
            return res

        with open(path, 'wb') as f:
            for chunk in res.iter_content(chunk_size=1024 * 32):
                if chunk:  # filter out keep-alive new chunks
                    f.write(chunk)

        return path

    def load_arrow_table(self, datastore_id: UUID, limit: Optional[int] = None, batch_size: int = 1024) -> \
            Union[DtlError, Table]:
        """
        Loads the resource in an in memory arrow buffer.
        This can be used either with Nvidia RAPIDS or pandas

        pandas:
        df = dtl.client.datastore.load_arrow_table(datastore_id).to_pandas()

        :param datastore_id: if of the resource to load
        :param limit: elements to be loaded in the frame
        :param batch_size: size of the batch to be retrieved at a time from the collection
        :return:
        """

        params = {
            "fileFormat": FileFormat.Csv.value
        }

        if limit is not None:
            if not isinstance(limit, int) or limit < 0:
                return DtlError("limit should be a positive integer")

            params["limit"] = limit

        res = self.http_client.make_authed_request(
            self.service_uri + "/datastores/" + str(datastore_id) + "/download",
            HttpMethod.GET, params=params, stream=True
        )

        if isinstance(res, DtlError):
            return res

        stream = ResponseStream(res.iter_content(batch_size))

        return csv.read_csv(stream)

    def upload(self, datastore_id: UUID, file_format: FileFormat, path: str, params: dict = {}) -> Union[DtlError, None]:
        """
        Enable to upload a file to the specified resource through the platform

        :param datastore_id: id of the resource to upload the data through
        :param file_format: format of the file to be uploaded
        :param path: location of the file to upload on disk
        :return:
        """

        with open(path, 'rb') as f:
            files = [MultiPartFile("content", f)]

            data = [
                MultiPartData("file-format", file_format.value),
                MultiPartData("params", "{}"),
            ]

            res = self.http_client.post_multipart_data(
                self.service_uri + "/datastores/" + str(datastore_id) + "/upload",
                files, data
            )

            if isinstance(res, DtlError):
                return res




import abc
from typing import Optional, Union
from datalogue.models.datastore import CredentialType, credential_type_from_str
from datalogue.errors import _enum_parse_error, DtlError
from datalogue.utils import SerializableStringEnum
from uuid import UUID


class CredentialsReference:
    def __init__(self, ref_id: UUID, name: Optional[str], ref_type: CredentialType):
        self.id = ref_id
        self.name = name
        self.type = ref_type

    def __eq__(self, other: 'CredentialsReference'):
        if isinstance(self, other.__class__):
            return self.id == other.id and self.name == other.name and self.type == other.type
        return False

    def __repr__(self):
        return f'{self.__class__.__name__}(id: {self.id}, name: {self.name!r}, type: {self.type!r})'


def _credentials_ref_from_payload(json: dict) -> Union[DtlError, CredentialsReference]:
    ref_id = json.get("id")
    if not isinstance(ref_id, str):
        return DtlError("A credentials reference needs an 'id' field")
    else:
        ref_id = UUID(ref_id)

    ref_type = json.get("type")
    if not isinstance(ref_type, str):
        return DtlError("A credentials reference needs a '_type' field")
    else:
        ref_type = credential_type_from_str(ref_type)

    name = json.get("name")
    return CredentialsReference(ref_id, name, ref_type)


class Credentials(abc.ABC):

    type_field = "type"

    def __init__(self, definition_type: CredentialType):
        self.type = definition_type

    def _base_payload(self) -> dict:
        return dict([(Credentials.type_field, self.type.value)])

    def __eq__(self, other: 'Credentials'):
        if isinstance(self, other.__class__):
            return self._as_payload() == other._as_payload()
        return False

    @abc.abstractmethod
    def _as_payload(self) -> dict:
        """
        Dictionary representation of the object
        :return:
        """


class S3Credentials(Credentials):
    type_str = CredentialType.S3

    def __init__(self, region: str, client_id: str, client_secret: str):
        Credentials.__init__(self, S3Credentials.type_str)
        self.region = region
        self.client_id = client_id
        self.client_secret = client_secret

    def __repr__(self):
        return f'{self.__class__.__name__}(region: {self.region!r}, client_id: {self.client_id!r}, client_secret: ****)'

    def _as_payload(self) -> dict:
        """
        Dictionary representation of the object
        :return:
        """
        base = self._base_payload()
        base["region"] = self.region
        base["clientId"] = self.client_id
        base["clientSecret"] = self.client_secret
        return base


class S3ACredentials(S3Credentials):
    type_str = CredentialType.Hadoop

    def __init__(self, region: str, client_id: str, client_secret: str):
        S3Credentials.__init__(self, region, client_id, client_secret)
        self.type = S3ACredentials.type_str


class AzureBlobType(SerializableStringEnum):
    Block = "Block"
    Append = "Append"

    @staticmethod
    def parse_error(s: str) -> str:
        return _enum_parse_error("azure blob type", s)

    @staticmethod
    def from_payload(json: str) -> Union[DtlError, 'AzureBlobType']:
        return SerializableStringEnum.from_str(AzureBlobType)(json)


class AzureEndpointProtocol(SerializableStringEnum):
    Https = "Https"
    Http = "Http"

    @staticmethod
    def parse_error(s: str) -> str:
        return _enum_parse_error("azure blob type", s)

    @staticmethod
    def from_payload(json: str) -> Union[DtlError, 'AzureBlobType']:
        return SerializableStringEnum.from_str(AzureEndpointProtocol)(json)


class AzureCredentials(Credentials):
    type_str = CredentialType.Azure

    def __init__(self, account_name: str, account_key: str, endpoint_protocol: AzureEndpointProtocol,
                 blob_type: AzureBlobType):
        Credentials.__init__(self, AzureCredentials.type_str)
        self.account_name = account_name
        self.account_key = account_key
        self.endpoint_protocol = endpoint_protocol
        self.blob_type = blob_type

    def __repr__(self):
        return f'{self.__class__.__name__}(account_name: {self.account_name!r},' \
               f' endpoint_protocol: {self.endpoint_protocol!r}, blob_type: {self.blob_type!r}, account_key: ****)'

    def _as_payload(self) -> dict:
        """
        Dictionary representation of the object
        :return:
        """
        base = self._base_payload()
        base["accountName"] = self.account_name
        base["accountKey"] = self.account_key
        base["endpointProtocol"] = self.endpoint_protocol.value
        base["blobType"] = self.blob_type.value
        return base


class GCS(Credentials):
    type_str = CredentialType.GCS

    def __init__(self, client_email: str, private_key: str):
        Credentials.__init__(self, GCS.type_str)
        self.client_email = client_email
        self.private_key = private_key

    def __repr__(self):
        return f'{self.__class__.__name__}(client_email: {self.client_email!r}, private_key: ****)'

    def _as_payload(self) -> dict:
        """
        Dictionary representation of the object
        :return:
        """
        base = self._base_payload()
        base["clientEmail"] = self.client_email
        base["privateKey"] = self.private_key
        return base


class JdbcCredentials(Credentials):
    type_str = CredentialType.JDBC

    def __init__(self, url: str, schema: str, user: str, password: str):
        Credentials.__init__(self, JdbcCredentials.type_str)
        self.url = url
        self.schema = schema
        self.user = user
        self.password = password

    def __repr__(self):
        return f'{self.__class__.__name__}(url: {self.url!r}, schema: {self.schema!r}, user: {self.user!r}, password: ****)'

    def _as_payload(self) -> dict:
        """
        Dictionary representation of the object
        :return:
        """
        base = self._base_payload()
        base["url"] = self.url
        base["schema"] = self.schema
        base["user"] = self.user
        base["password"] = self.password
        return base


class MongoCredentials(Credentials):
    type_str = CredentialType.Mongo

    def __init__(self, url: str, database: str, user: str, password: str):
        Credentials.__init__(self, MongoCredentials.type_str)
        self.url = url
        self.database = database
        self.user = user
        self.password = password

    def __repr__(self):
        return f'{self.__class__.__name__}(url: {self.url!r}, database: {self.database!r}, user: {self.user!r}, password: ****)'

    def _as_payload(self) -> dict:
        """
        Dictionary representation of the object
        :return:
        """
        base = self._base_payload()
        base["url"] = self.url
        base["database"] = self.database
        base["user"] = self.user
        base["password"] = self.password
        return base


class SocrataCredentials(Credentials):
    type_str = CredentialType.Socrata

    def __init__(self, domain: str, token: str):
        Credentials.__init__(self, SocrataCredentials.type_str)
        self.domain = domain
        self.token = token

    def __repr__(self):
        return f'{self.__class__.__name__}(domain: {self.domain!r}, token: ****)'

    def _as_payload(self) -> dict:
        """
        Dictionary representation of the object
        :return:
        """
        base = self._base_payload()
        base["domain"] = self.domain
        base["token"] = self.token
        return base

class KinesisCredentials(Credentials):
    type_str = CredentialType.Kinesis

    def __init__(self, region: str, client_id: str, client_secret: str):
        Credentials.__init__(self, KinesisCredentials.type_str)
        self.region = region
        self.client_id = client_id
        self.client_secret = client_secret

    def __repr__(self):
        return f'{self.__class__.__name__}(region: {self.region!r}, client_id: {self.client_id!r}, client_secret: ****)'
    
    def _as_payload(self):
        """
        Dictionary representation of the object
        :return:
        """
        base = self._base_payload()
        base["region"] = self.region
        base["clientId"] = self.client_id
        base["clientSecret"] = self.client_secret
        return base

class KafkaCredentials(Credentials):
    type_str = CredentialType.Kafka

    def __init__(self, bootstrap_server: str):
        Credentials.__init__(self, KafkaCredentials.type_str)
        self.bootstrap_server = bootstrap_server

    def __repr__(self):
        return f'{self.__class__.__name__}(bootstrap_server: {self.bootstrap_server!r})'

    def _as_payload(self):
        """
        Dictionary representation of the object
        :return:
        """
        base = self._base_payload()
        base["bootstrapServer"] = self.bootstrap_server
        return base


from typing import List, Union, Optional
from datalogue.clients.http import _HttpClient, HttpMethod
from datalogue.models.stream import Stream
from datalogue.models.stream_collection import StreamCollection, _stream_collection_from_payload, _stream_metadata_from_payload
from datalogue.utils import _parse_list
from datalogue.models.job import _job_from_payload
from datalogue.errors import DtlError
from uuid import UUID, uuid4
from datetime import datetime, timedelta
from dateutil.tz import UTC


class _StreamCollectionClient:
    """
    Right now interactions on Stream Collections as they are currently called in the API. Or Stream Collection as they are
    called in the UI.

    To simplify things, right now the SDK allows interactions with Stream Collection of one stream
    """

    def __init__(self, http_client: _HttpClient):
        self.http_client = http_client
        self.service_uri = "/scout"

    def create(self, streams: List[Stream], name: Optional[str]) -> Union[DtlError, StreamCollection]:
        """
        Creates the stream as specified.
        Right now this creates a Stream Collection with only one stream

        TODO think about creating datastores when they are references in the Stream that don't exist

        :param streams: Streams to be created
        :param name: name of the stream to be created. If None supplied, one will be generated
        :return: string with error message if failed, uuid otherwise
        """

        if name is None:
            name = "stream-" + str(uuid4())[0:8]

        # save Stream Collection
        res1 = self.http_client.make_authed_request(
            self.service_uri + '/stream-collections', HttpMethod.POST, { "name": name }
        )

        stream_collection_id = res1.get("id")
        if stream_collection_id is None:
            return DtlError("Created response did not return an id")

        created_streams = []
        # add stream
        for stream in streams:
            res2 = self.http_client.make_authed_request(
                self.service_uri + '/stream-collections/' + stream_collection_id + "/stream", HttpMethod.POST, stream._as_payload()
            )

            if isinstance(res2, DtlError):
                return res2
            
            stream_meta = _stream_metadata_from_payload(res2)
            if isinstance(stream_meta, DtlError):
                return stream_meta

            created_streams.append(stream_meta)

        return StreamCollection(stream_collection_id, name, created_streams)

    def update(self, stream_collection_id: UUID, name: Optional[str] = None, streams: Optional[List[Stream]] = None) \
            -> Union[DtlError, StreamCollection]:
        """
        Updates the Stream Collection

        :param stream_collection_id: id of the Stream Collection to update
        :param name: new name of the Stream Collection
        :param streams: new set of streams for the Stream Collection (overrides the previous set)
        :return: Either an error message in a string or the Stream Collection
        """

        ref_stream_collection = self.get(stream_collection_id)
        if isinstance(ref_stream_collection, DtlError):
            return DtlError("It looks like you are trying to update a Stream Collection that doesn't exist", ref_stream_collection)

        if isinstance(name, str):

            res = self.http_client.make_authed_request(
                f"{self.service_uri}/stream-collections/{stream_collection_id}", HttpMethod.PUT, {"name": name}
            )

            if isinstance(res, DtlError):
                return res

        if isinstance(streams, list):

            # We start by removing all the existing streams
            for stream in ref_stream_collection.streams:
                res = self.http_client.make_authed_request(f"{self.service_uri}/streams/{stream.id}", HttpMethod.DELETE)

                if isinstance(res, DtlError):
                    return DtlError("Could not delete all previously existing streams, interrupting update", res)

            # Then we attach all the new ones
            for stream in streams:
                res = self.http_client.make_authed_request(
                    f"{self.service_uri}/stream-collections/{stream_collection_id}/stream", HttpMethod.POST, stream._as_payload()
                )
                if isinstance(res, DtlError):
                    return DtlError("Could not add all the new streams to the Stream Collection, interrupting update", res)

        return self.get(stream_collection_id)

    def schedule(self, stream_collection_id: UUID, date: datetime) -> Union[DtlError, bool]:
        """
        Creates a job to be run at the given date

        :param stream_collection_id: id of the stream to be run
        :param date: date at which to run the stream, expected to be localized
        :return: Either an error message in a string or the UUID for the new job
        """

        res = self.http_client.make_authed_request(
            self.service_uri + '/stream-collections/' + str(stream_collection_id) + "/schedules",
            HttpMethod.POST,
            {"runDate": date.astimezone(UTC).strftime("%Y-%m-%dT%H:%M:%SZ")}
        )

        if isinstance(res, DtlError):
            return res

        return _parse_list(_job_from_payload)(res)

    def run(self, pipeline_id: UUID) -> Union[DtlError, bool]:
        """
        Runs the pipeline right now + 2 secs

        :param pipeline_id: id of the pipeline to run
        :return:
        """

        return self.schedule(pipeline_id, datetime.now() + timedelta(seconds=2))

    def list(self, page: int = 1, item_per_page: int = 25) -> Union[DtlError, List[StreamCollection]]:
        """
        Retrieves a list of the available streams (Stream Collections)

        TODO pagination
        :param page: page to be retrieved
        :param item_per_page: number of items to be put in a page
        :return: Returns a List of all the available streams or an error message as a string
        """
        res = self.http_client.make_authed_request(self.service_uri + "/stream-collections", HttpMethod.GET)

        if isinstance(res, DtlError):
            return res

        return _parse_list(_stream_collection_from_payload)(res)

    def get(self, stream_collection_id: UUID) -> Union[DtlError, StreamCollection]:
        """
        From the provided id, get the corresponding stream (Stream Collection)

        :param stream_collection_id:
        :return:
        """
        res = self.http_client.make_authed_request(
            self.service_uri + "/stream-collections/" + str(stream_collection_id), HttpMethod.GET)

        if isinstance(res, DtlError):
            return res

        return _stream_collection_from_payload(res)

    def delete(self, stream_collection_id: UUID) -> Union[DtlError, bool]:
        """
        Deletes the given stream (Stream Collection)

        :param Stream Collection_id: id of the stream (Stream Collection) to be deleted
        :return: true if successful, false otherwise
        """

        res = self.http_client.make_authed_request(
            self.service_uri + "/stream-collections/" + str(stream_collection_id), HttpMethod.DELETE)

        if isinstance(res, DtlError):
            return res
        else:
            return True



from typing import *
from typing import BinaryIO
from enum import Enum
import os
from datalogue.utils import Json
from datalogue.errors import DtlError
import logging
import requests
from requests.packages.urllib3.exceptions import InsecureRequestWarning
from requests import Response
from datalogue.models.organization import _users_from_payload

requests.packages.urllib3.disable_warnings(InsecureRequestWarning)

logger = logging.getLogger("_HttpClient")


class HttpMethod(Enum):
    POST = 'POST'
    GET = 'GET'
    PUT = 'PUT'
    DELETE = 'DELETE'


class MultiPartFile:
    def __init__(self, name: str, file: BinaryIO):
        self.name = name
        self.file = file

class MultiPartData:
    def __init__(self, name: str, data: str):
        self.name = name
        self.data = data

def _has_ssl_certificate():
    """
    Check if SSL certificate is installed, prints to logger if not verified, and returns the certificate.
    :return: True if contains SSL certificate, False otherwise
    """
    verify_cert_env = os.environ.get('DTL_SSL_VERIFY_CERT')
    verify_certificate = False if (verify_cert_env == 'false') else True
    if not verify_certificate:
        logger.info('SSL Certification verification is disabled.')
    return verify_certificate


class _HttpClient:
    """
    Class that Abstracts aways requests and makes it use type safe.
    It also prepares all the requests to be used authenticated to the API
    """

    def __init__(self, uri: str):
        self.uri = uri
        self.http_session = requests.Session()

    def login(self, username: str, password: str) -> Union[DtlError, None]:
        """
        Sets the cookie for the client to use

        :return: Nothing
        """
        verify_certificate = _has_ssl_certificate()

        url = self.uri + '/signin'
        try:
            res = self.http_session.post(
                url,
                json={
                    'email': username,
                    'password': password
                },
                allow_redirects=False,
                verify=verify_certificate
            )
        except requests.ConnectionError:
            return DtlError("There was a connection error trying to reach: %s" % self.uri)
        except requests.HTTPError:
            return DtlError("There was an HTTP error trying to reach: %s" % self.uri)
        except requests.URLRequired:
            return DtlError("%s is not a valid endpoint." % self.uri)
        except requests.Timeout:
            return DtlError("The request timed out")

        if res.status_code != 201 and res.status_code != 302 and res.status_code != 200:
            return DtlError('Authentication failed - %s : %s' % (str(res.status_code), res.text))
        elif res.status_code == 503:
            return DtlError("The service is not available")
        else:
            return None

    def signup(self, first_name, last_name, email, password, accept_terms):
        """
        Allows
        :param first_name: First name of user to be registered
        :param last_name: Last name of user to be registered
        :param email: Email of user to be registered
        :param password: Password of user to be registered
        :param accept_terms: Whether user accepts the term of service here : https://www.datalogue.io/pages/terms-of-service
        :return: A User object
        """
        print("By calling this action, you've agreed to the terms of service as set forth here : "
                   "https://www.datalogue.io/pages/terms-of-service")
        verify_certificate = _has_ssl_certificate()

        if not accept_terms:
            return DtlError("Terms & Conditions for registration are not accepted")
        url = self.uri + '/signup'
        try:
            res = self.http_session.post(url, json={
                    'email': email,
                    'firstName': first_name,
                    'lastName': last_name,
                    'password': password
                }, verify=verify_certificate)
        except requests.ConnectionError:
            return DtlError("There was a connection error trying to reach: %s" % self.uri)
        except requests.HTTPError:
            return DtlError("There was an HTTP error trying to reach: %s" % self.uri)
        except requests.URLRequired:
            return DtlError("%s is not a valid endpoint." % self.uri)
        except requests.Timeout:
            return DtlError("The request timed out")

        if res.status_code == 200:
            if res.headers.get("content-type") == "application/json":
                return _users_from_payload(res.json())
            else:
                return DtlError("Cannot parse response, unknown type")
        elif res.status_code == 503:
            return DtlError("The service is not available")
        else:
            return DtlError(res.text)

    @staticmethod
    def _handle_status_code(request_uri: str, res: Response) ->  Union[DtlError, None]:
        """
        Return a proper DtlError depending on the status code of the Response

        :param request_uri: uri of th request that was made
        :param res: Response received to the request
        :return: an Error or None if the status doesn't warrant returning an error
        """
        if res.status_code == 400:
            return DtlError("Invalid Request to %s returned 400, body res: %s" % (request_uri, res.text))

        if res.status_code == 401:
            return DtlError("It looks like you are not logged in, please login")

        if res.status_code == 403:
            return DtlError(f"Authorization error. Make sure you're authorized to do this operation. Msg: {res.text}")

        if res.status_code == 404:
            return DtlError("Route not found %s " % request_uri)

        if res.status_code == 413:
            return DtlError("Body is too big, please limit uploads to 1 Go")

        if res.status_code == 500:
            return DtlError("Internal Server error")

        if res.status_code == 503:
            return DtlError("The service is not available")

        return None

    def post_multipart_data(self, path: str, files: List[MultiPartFile], data: List[MultiPartData]) -> Union[DtlError, Response]:
        headers = { 'content-type': "multipart/form-data" }
        
        files_payload = {}
        for f in files:
            files_payload[f.name] = f.file

        data_payload = {}
        for d in data:
            data_payload[d.name] = d.data
        
        request_uri = self.uri + path
        req = requests.Request(
            HttpMethod.POST.value, 
            request_uri, 
            files=files_payload, 
            data=data_payload)
        prepped = self.http_session.prepare_request(req)
        res = self.http_session.send(prepped)

        e = self. _handle_status_code(request_uri, res)
        if isinstance(e, DtlError):
            return e
        
        content_type_header = res.headers.get("content-type")
        if len(res.text) > 0 and content_type_header == "application/json":
            return res.json()
        else:
            return res.text

    def post_binary(self, path: str, file: BinaryIO, params: Optional[dict] = None) -> Union[DtlError, Response]:
        """
        Makes a post request with the content being a binary file that will be uploaded chunk by chunk

        :param path: path uri to the datastore
        :param file: iterator that is going to generate the binary data
        :param params: parameters to be used in the request
        :return:
        """
        request_uri = self.uri + path
        req = requests.Request(HttpMethod.POST.value, request_uri, data=file, params=params)
        prepped = self.http_session.prepare_request(req)

        prepped.headers['Content-Type'] = 'application/octet-stream'

        res = self.http_session.send(prepped)

        e = _HttpClient._handle_status_code(request_uri, res)
        if isinstance(e, DtlError):
            return e

    def make_authed_request(self, path: str, method: HttpMethod, body: Optional[dict] = None,
                            params: Optional[dict] = None, stream=False, verify=True) -> Union[DtlError, Json, Response]:
        """
        Makes a request to the API give the url

        :param path: path to access the datastore below the host
        :param method: parameter to indicate HTTP Method to be used in the request
        :param body: body to be sent along with the request
        :param params: query parameters, to be fed with a dictionary
        :param stream: boolean to activate streaming the content of the response, this will delay the retrieval of
        the body until .content is called.
        :return: Request response in either raw format, json or text
        """
        request_uri = self.uri + path
        req = requests.Request(method.value, request_uri, json=body, params=params)
        prepped = self.http_session.prepare_request(req)
        res = self.http_session.send(prepped, stream=stream, verify=verify)

        e = self. _handle_status_code(request_uri, res)
        if isinstance(e, DtlError):
            return e

        if method == HttpMethod.DELETE:
            return True
        else:
            content_type_header = res.headers.get("content-type")
            if stream:
                return res
            elif len(res.text) > 0 and content_type_header == "application/json":
                return res.json()
            else:
                return res.text



from typing import List, Union
from datalogue.clients.http import _HttpClient, HttpMethod
from datalogue.models.datastore_collection import DatastoreCollection, _datastore_collection_from_dict
from datalogue.errors import DtlError
from uuid import UUID
from datalogue.utils import _parse_list


class _DatastoreCollectionClient:
    """
    Client to interact with the datastore collection objects
    """

    def __init__(self, http_client: _HttpClient):
        self.http_client = http_client
        self.service_uri = "/scout"

    def create(self, datastore_collection: DatastoreCollection) -> Union[DtlError, DatastoreCollection]:
        """
        Creates the Datastore Collection as specified.

        :param datastore_collection: Datastore Collection to be created
        :return: string with error message if failed, uuid otherwise
        """
        assert isinstance(datastore_collection, DatastoreCollection), 'Input is not a Datastore Collection instance.'
        # todo?? assert datastore indeed is a datastore

        res = self.http_client.make_authed_request(
            self.service_uri + "/datastore-collections", HttpMethod.POST,
            datastore_collection._as_payload())

        if isinstance(res, DtlError):
            return res

        return _datastore_collection_from_dict(res)

    def update(self, datastore_collection_id: UUID, datastore_collection: DatastoreCollection) -> Union[DtlError, bool]:
        """
        Updates the Datastore Collection for the given datastore_collection_id

        :param datastore_collection_id: Id of the datastore collection to update
        :param datastore_collection: Datastore Collection to be updated
        :return: string with error message if failed, uuid otherwise
        """
        assert isinstance(datastore_collection, DatastoreCollection), 'Input is not a Datastore Collection instance.'

        res = self.http_client.make_authed_request(
            self.service_uri + "/datastore-collections/" + str(datastore_collection_id), HttpMethod.PUT,
            datastore_collection._as_payload())


        if isinstance(res, DtlError):
            return res
        else:
            return True

    def search(self, query: str, page: int = 1, item_per_page: int = 25) -> Union[DtlError, List[DatastoreCollection]]:
        """
        Search through the Datastore Collections

        :param query: text query to be used to make the search
        :param page: page to be retrieved
        :param item_per_page: number of items to be put in a page
        :return: Returns a List of all the available streams or an error message as a string
        """

        assert isinstance(query, str), 'Input query must be a string.'

        res = self.http_client.make_authed_request(
            self.service_uri + "/datastore-collections?", HttpMethod.GET, )

        if isinstance(res, DtlError):
            return res

        out_list = []

        for i in range(len(res)):
            for content in res[i].values():
                if query in str(content):
                    out_list.append(res[i])
                    break

        # Todo build pagination
        return _parse_list(_datastore_collection_from_dict)(out_list)

    def list(self, page: int = 1, item_per_page: int = 25) -> Union[DtlError, List[DatastoreCollection]]:
        """
        List the Datastore Collection

        :param page: page to be retrieved
        :param item_per_page: number of items to be put in a page
        :return: Returns a List of all the available datastore collections or an error message as a string
        """
        res = self.http_client.make_authed_request(
            self.service_uri + "/datastore-collections?", HttpMethod.GET, )

        if isinstance(res, DtlError):
            return res

        # Todo build pagination
        return _parse_list(_datastore_collection_from_dict)(res)

    def get(self, datastore_collection_id: UUID) -> Union[DtlError, DatastoreCollection]:
        """
        From the provided id, get the corresponding Datastore Collection

        :param datastore_collection_id:
        :return:
        """
        res = self.http_client.make_authed_request(
            self.service_uri + "/datastore-collections/" + str(datastore_collection_id),
            HttpMethod.GET)

        if isinstance(res, DtlError):
            return res

        return _datastore_collection_from_dict(res)

    def delete(self, datastore_collection_id: UUID) -> Union[DtlError, bool]:
        """
        Deletes the given Datastore Collection

        :param datastore-collection_id: id of the datastore collection to be deleted
        :return: true if successful, the error otherwise
        """
        res = self.http_client.make_authed_request(
            self.service_uri + "/datastore-collections/" + str(datastore_collection_id),
            HttpMethod.DELETE
        )

        if isinstance(res, DtlError):
            return res
        else:
            return True



from typing import Optional, Union, List
from datalogue.clients.http import _HttpClient, HttpMethod
from datalogue.models.ontology import Ontology, OntologyNode, DataRef
from datalogue.models.training import Training, ModelType, TrainingStatusType, OrderList, TrainingState, training_status_type_from_str
from datalogue.errors import DtlError
from datalogue.models.organization import User
from datalogue.utils import _parse_list
from uuid import UUID
from datalogue.clients.user import _UserClient

class _DataClient:
    """
    Client to interact with TrainingData
    """
    def __init__(self, http_client: _HttpClient):
        self.http_client = http_client

    def add(self, store_id: UUID, store_name: str, refs: List[DataRef]) -> Union[DtlError, List[UUID]]:
        """
        Attaches paths in the given `store_id` to the the nodes of the ontology.

        :param store_id: Id of the datastore that it going to be read
        :param store_name: Name of the datastore that it going to be read
        :param refs: List of data references to show which paths are going to be attached which ontology nodes
        :return: List of stream ids which are the jobs that is transferring data from datastore to Themis
        """
        stream_ids = []

        for dataRef in refs:
            for path in dataRef.path_list:

                # Creates a dataset per attachment as UI does.
                dataset_id = self.__create_dataset(store_name)
                if isinstance(dataset_id, DtlError):
                    return dataset_id
                
                stream_id = self.__transfer_data_from_datastore(store_id, dataset_id, dataRef.node.node_id, path)
                if isinstance(dataset_id, DtlError):
                    return stream_id

                self.__update_node(dataRef.node.node_id, path, dataset_id, stream_id)
                stream_ids.append(stream_id)
        return stream_ids

    def __create_dataset(self, store_name: str) -> Union[DtlError, UUID]:
        payload = {
            "title": store_name,
            "tags": [],
            "label_map": {}
        }

        res = self.http_client.make_authed_request('/themis/dataset', HttpMethod.POST, body = payload)
        dataset_id = UUID(res.get("id"))

        if dataset_id is None:
            return DtlError("There is no dataset id in response!")
        return dataset_id

    def __transfer_data_from_datastore(self, store_id: UUID, dataset_id: UUID, node_id: UUID, path: List[str]) -> UUID:
        path_param = "&".join(map(lambda s: f"path={s}", path))
        url = f"/scout/run/training-data?sourceId={store_id}&{path_param}&trainingDataId={dataset_id}&class={str(node_id)}"

        res = self.http_client.make_authed_request(url, HttpMethod.POST)

        stream_id = UUID(res["streamId"])
        return stream_id

    def __update_node(self, node_id: UUID, path: List[str], dataset_id: UUID, stream_id: UUID) -> None:
        #TODO We should remove this part when we move training data adding functionality out of Yggy
        entity_res = self.http_client.make_authed_request(f"/yggy/entity/{node_id}", HttpMethod.GET)
        training_data_list = entity_res.get("trainingDataInfo")
        if training_data_list is None:
            training_data_list = []

        training_data_list.append({
            "datasetId": str(dataset_id),
            "nodePath": path,
            "streamId": str(stream_id)
        })

        self.http_client.make_authed_request(f"/yggy/entity/{node_id}", HttpMethod.POST,
            body={"trainingDataInfo": training_data_list})
        

class _TrainingClient:
    """
    Client to interact with the Trainings
    """
    def __init__(self, http_client: _HttpClient):
        self.http_client = http_client
        self.data = _DataClient(http_client)

    def deploy(self, training_id: UUID, ontology_id: UUID) -> Union[DtlError, bool]:
        """
        Deploy :class:`Training` based on the given training_id and ontology_id
        :return: True if successful, else returns :class:`DtlError`
        """
        res = self.http_client.make_authed_request(
            f'/yggy/ontology/{str(ontology_id)}/trainings/{str(training_id)}/deploy', 
            HttpMethod.POST, {})

        if isinstance(res, DtlError):
            return res
        else:
            return True

    #TODO API is not returning a training object, we should change this when the API returns training
    def run(self, ontology_id: UUID) -> Union[DtlError, bool]:
        """
        Starts training for a given ontology_id

        :param ontology_id:
        :return: Either a :class:`DtlError` if any error occurs or :class:`True` if training is requested successfully.
        """
        res = self.http_client.make_authed_request(f"/yggy/ontology/{ontology_id}/train", HttpMethod.POST)

        if isinstance(res, DtlError):
            return res
        return True

    def get_trainings(self, ontology_id: UUID, model_type: Optional[ModelType] = None, status_type: Optional[TrainingStatusType] = None, sort: Optional[OrderList] = OrderList.desc, limit: Optional[int] = 20) -> Union[DtlError, List[TrainingState]]:
        """
        Get :class:`Trainings` based on the given ontology_id
        
        :param ontology_id:
        :return: List of trainings if successful, else returns :class:`DtlError`
        """

        params = {}

        if limit is not None:
            if not isinstance(limit, int) or limit < 0:
                return DtlError("limit should be a positive integer")
            params["limit"] = limit

        if model_type is not None:
            params["model-type"] = model_type.value
        
        if status_type is not None:
            params["status-type"] = status_type.value
        
        if sort is not None:
            params["sort"] = sort.value
        
        res = self.http_client.make_authed_request(
            f'/yggy/ontology/{str(ontology_id)}/trainings',
            HttpMethod.GET, params=params)
        
        if isinstance(res, DtlError):
            return res
        else:
            return _parse_list(TrainingState._from_payload)(res)

    def cancel(self, training_id: UUID) -> Union[DtlError, bool]:
        """
        Cancel training given a training_id

        :param training_id:
        :return: status of the training if successful, else returns :class:`DtlError`
        """

        res = self.http_client.make_authed_request(
            f'/argus/trainings/{str(training_id)}/cancel',
            HttpMethod.POST)

        if isinstance(res, DtlError):
            return res
        else:
            return True


def get_training_history(self, ontology_id: UUID, limit: Optional[int] = 20) -> Union[DtlError, List[Training]]:
        """
        Get :class:`Training` history based on the given ontology_id
        
        :param ontology_id:
        :return: List of trainings if successful, else returns :class:`DtlError`
        """

        params = {}
        if limit is not None:
            if not isinstance(limit, int) or limit < 0:
                return DtlError("limit should be a positive integer")
            params["limit"] = limit
        
        res = self.http_client.make_authed_request(
            f'/yggy/ontology/{str(ontology_id)}/trainings/history',
            HttpMethod.GET, params=params)
        
        if isinstance(res, DtlError):
            return res
        else:
           return _parse_list(Training._from_payload)(res)





from datetime import datetime
from uuid import UUID
from typing import Union, Optional

from dateutil.parser import parse

from datalogue.utils import SerializableStringEnum
from datalogue.errors import _enum_parse_error, DtlError


class JobStatus(SerializableStringEnum):
    Scheduled = "Scheduled"
    Defined = "Defined"
    Running = "Running"
    Succeeded = "Succeeded"
    Failed = "Failed"
    Unknown = "Unknown"
    Cancelled = "Cancelled"

    @staticmethod
    def parse_error(s: str) -> str:
        return _enum_parse_error("job status", s)


def job_status_from_str(string: str) -> Union[DtlError, JobStatus]:
    return SerializableStringEnum.from_str(JobStatus)(string)


class Job:
    def __init__(self, stream_id: UUID, run_at: datetime, status: JobStatus, stream_collection_id: UUID, job_id: UUID, remaining_time_millis: int, percentage_progress: int, created_by: UUID, details: str, ended_at: Optional[datetime]):
        self.stream_id = stream_id
        self.stream_collection_id = stream_collection_id
        self.status = status
        self.run_at = run_at
        self.job_id = job_id
        self.remaining_time_millis = remaining_time_millis
        self.percentage_progress = percentage_progress
        self.created_by = created_by
        self.details = details
        self.ended_at = ended_at

    def __repr__(self):
        return (f'{self.__class__.__name__}('
                f'job_id: {self.job_id!r}, '
                f'stream_id: {self.stream_id!r}, '
                f'stream_collection_id: {self.stream_collection_id!r}, '
                f'status: {self.status!r}, '
                f'run_at: {self.run_at!r}, '
                f'created_by: {self.created_by!r}, '
                f'remaining_time_millis: {self.remaining_time_millis!r}, '
                f'percent_progress: {self.percentage_progress!r}, '
                f'errors: {self.details!r}), '
                f'ended_at: {self.ended_at!r}')


def _job_from_payload(json: dict) -> Union[DtlError, Job]:
    stream_id = json.get('streamId')
    if stream_id is None:
        return DtlError("Job object should have a 'stream_id' property")
    else:
        try:
            stream_id = UUID(stream_id)
        except ValueError:
            return DtlError("'streamId' field was not a proper uuid")

    run_at = json.get("runDate")
    if run_at is None:
        return DtlError("Job object should have a 'runDate' property")
    else:
        try:
            run_at = parse(run_at)
        except ValueError:
            return DtlError("The 'runDate' could not be parsed as a valid date")

    status = json.get("status")
    if status is None:
        return DtlError("Job object should have a 'status' property")
    else:
        status = job_status_from_str(status)
        if isinstance(status, DtlError):
            return status

    stream_collection_id = json.get("streamCollectionId")
    if stream_collection_id is None:
        return DtlError("Job object should have a 'streamCollectionId' property")
    else:
        try:
            stream_collection_id = UUID(stream_collection_id)
        except ValueError:
            return DtlError("'streamCollectionId' field was not a proper uuid")

    job_id = json.get('jobId')
    if job_id is None:
        return DtlError("Job object should have a 'jobId' property")
    else:
        try:
            job_id = UUID(job_id)
        except ValueError:
            return DtlError("'jobId' field was not a proper uuid")

    remaining_time_millis = json.get('remainingTimeInMillis')
    if remaining_time_millis is None:
        return DtlError("Job object should have a 'remainingTimeInMillis' property")
    else:
        try:
            remaining_time_millis = int(remaining_time_millis)
        except ValueError:
            return DtlError("'remainingTimeInMillis' was not a proper int")

    percentage_progress = json.get('percentage')
    if percentage_progress is None:
        return DtlError("Job object should have a 'percentProgress' property")
    else:
        try:
            percentage_progress = int(percentage_progress)
        except ValueError:
            return DtlError("'percentProgress' was not a proper int")

    created_by = json.get('createdBy')
    if created_by is None:
        return DtlError("Job object should have a 'createdBy' property")
    else:
        try:
            created_by = UUID(created_by)
        except ValueError:
            return DtlError("'createdBy' was not a proper uuid")

    # details is optional
    details = json.get('details')

    #ended_at is optional
    ended_at = json.get("endDate")
    if ended_at is not None:
        try:
            ended_at = parse(ended_at)
        except ValueError:
            return DtlError("The 'endDate' could not be parsed as a valid date")

    return Job(stream_id, run_at, status, stream_collection_id, job_id,
               remaining_time_millis, percentage_progress, created_by, details, ended_at)



from datalogue.models.transformations.commons import Transformation
from datalogue.utils import _parse_string_list, SerializableStringEnum, _parse_list
from datalogue.errors import _enum_parse_error, DtlError, _property_not_found
from typing import List, Union


class Function(SerializableStringEnum):
    """
    Functions that can be used to classify the context
    """
    EditDistance = "EditDistance"

    @staticmethod
    def parse_error(s: str) -> str:
        return _enum_parse_error("ClassifyContextFunction", s)

    @staticmethod
    def from_str(string: str) -> Union[DtlError, 'Function']:
        return SerializableStringEnum.from_str(Function)(string)


class ClassMapping:
    """
    Mapping between a class, references and a threshold
    """

    def __init__(self, tag: str, refs: List[str], threshold: int):
        """

        :param tag: class to be used as a tag
        :param refs: list of reference to compare the context to
        :param threshold: maximum distance allowed between the context and the reference for the datapoint to be
         classified as tag
        """
        self.tag = tag
        self.refs = refs
        self.threshold = threshold

    def __eq__(self, other: 'ClassMapping'):
        if isinstance(self, other.__class__):
            return self._as_payload() == other._as_payload()
        return False

    def __repr__(self):
        return f"ClassMapping(tag: {self.tag!r}, refs: {self.refs}, threshold: {self.threshold})"

    def _as_payload(self) -> dict:
        return {
            "type": "ClassMapping",
            "class": self.tag,
            "refs": self.refs,
            "threshold": self.threshold
        }

    @staticmethod
    def _from_payload(json: dict) -> Union[DtlError, 'ClassMapping']:
        t = json.get("type")
        if t != "ClassMapping":
            return DtlError("type property is not ClassMapping")

        tag = json.get("class")
        if tag is None:
            return _property_not_found("class", json)

        refs = json.get("refs")
        if refs is None:
            return _property_not_found("refs", json)
        else:
            refs = _parse_string_list(refs)
            if isinstance(refs, DtlError):
                return refs

        threshold = json.get("threshold")
        if threshold is None:
            return _property_not_found("threshold", json)

        return ClassMapping(tag, refs, threshold)


class ClassifyContext(Transformation):
    """
    ClassifyContext: Runs a function on the context (label) to be classify the datapoint
    """

    type_str = "ClassifyContext"

    def __init__(self, use: Function, options: List[ClassMapping]):
        """
        :param use: Function to be used to do the classification
        """
        Transformation.__init__(self, ClassifyContext.type_str)
        self.use = use.value
        self.options = options

    def __eq__(self, other: 'ClassifyContext'):
        if isinstance(self, other.__class__):
            return self._as_payload() == other._as_payload()
        return False

    def __repr__(self):
        return f"ClassifyContext(use: {self.use!r}, options: {self.options})"

    def _as_payload(self) -> dict:
        base = self._base_payload()
        base["use"] = self.use
        base["options"] = list(map(lambda o: o._as_payload(), self.options))
        return base

    @staticmethod
    def _from_payload(json: dict) -> Union[DtlError, 'ClassifyContext']:
        t = json.get("type")
        if t != "ClassifyContext":
            return DtlError("type property is not ClassifyContext")

        use = json.get("use")
        if use is None:
            return _property_not_found("use", json)

        use = Function.from_str(use)
        if isinstance(use, DtlError):
            return use

        options = json.get("options")
        if options is None:
            return _property_not_found("options", json)

        options = _parse_list(ClassMapping._from_payload)(options)
        if isinstance(options, DtlError):
            return options

        return ClassifyContext(use, options)



from typing import List, Union
from datalogue.errors import DtlError
from datalogue.models.transformations.commons import Transformation


class FlatMap(Transformation):
    """
    Creates multiple ADGs from one ADG and put each child into the given path of the new ADG.

    """
    type_str = "FlatMap"

    def __init__(self, on: List[str]):
        """
        :param on: array of string that represents the path where to put created ADGs 
        """
        Transformation.__init__(self, FlatMap.type_str)
        self.on = on

    def __eq__(self, other: 'FlatMap'):
        if isinstance(self, other.__class__):
            return self._as_payload() == other._as_payload()
        return False

    def __repr__(self):
        return f"FlatMap(on: {'.'.join(self.on)})"

    def _as_payload(self) -> dict:
        base = self._base_payload()
        base["on"] = self.on
        return base

    @staticmethod
    def _from_payload(json: dict) -> Union[DtlError, 'FlatMap']:
        on = json.get("on")
        if isinstance(on, DtlError):
            return on

        return FlatMap(on)



from typing import List, Union
from datalogue.clients.http import _HttpClient, HttpMethod
from datalogue.models.job import Job, _job_from_payload
from datalogue.utils import _parse_list
from datalogue.errors import DtlError
from uuid import UUID

class _JobsClient:
    """
    Client to interact with the Scheduled pipelines
    """

    def __init__(self, http_client: _HttpClient):
        self.http_client = http_client
        self.service_uri = "/scout"

    def list(self, page: int = 1, item_per_page: int = 25) -> Union[DtlError, List[Job]]:
        """
        List the scheduled jobs

        TODO Pagination

        :param page: page to be retrieved
        :param item_per_page: number of items to be put in a page
        :return: Returns a List of all the available Datastores or an error message as a string
        """
        res = self.http_client.make_authed_request(
            self.service_uri + '/stream-collections/schedules', HttpMethod.GET,
        )

        if isinstance(res, DtlError):
            return res

        return _parse_list(_job_from_payload)(res)

    def cancel(self, job_id: UUID) -> Union[DtlError, Job]:
        """"
        Cancel a Job given job_id.

        :param job_id:
        :return: Returns the Job object canceled or an error
        """

        res = self.http_client.make_authed_request(
            self.service_uri + f"/v2/jobs/{str(job_id)}/cancel", HttpMethod.POST,
        )

        if isinstance(res, DtlError):
            return res

        return _job_from_payload(res)



from typing import List, Union
from datalogue.errors import DtlError, _property_not_found, _invalid_property_type
from datetime import datetime
from datalogue.models.transformations.commons import Transformation, DataType
from datalogue.utils import _parse_list


class NodeDescription:
    """
    Description of a node to be added to the tree.
    """
    def __init__(self, path: List[str], value: Union[str, bool, int, float, datetime], value_type: DataType):
        """
        :param path: path of the node to be created
        :param value: value to be set for the node
        :param value_type: type of the value to be used
        """
        self.path = path
        self.value = value
        self.value_type = value_type

    def __eq__(self, other: 'NodeDescription'):
        if isinstance(self, other.__class__):
            return self._as_payload() == other._as_payload()
        return False

    def __repr__(self):
        return f"NodeDescription(path: {'.'.join(self.path)}, value: {self.value!r}, type: {self.value_type!r})"

    def _as_payload(self) -> Union[DtlError, dict]:
        value = self.value
        if isinstance(value, datetime):
            value = value.isoformat()
        else:
            value = str(value)

        return {
            "path": self.path,
            "value": value,
            "valueType": self.value_type.value
        }

    @staticmethod
    def _from_payload(json: dict) -> Union[DtlError, 'NodeDescription']:
        path = json.get("path")
        if path is None:
            return _property_not_found("path", json)
        elif not isinstance(path, list):
            return _invalid_property_type("path", "List[str]", json)

        value = json.get("value")
        if value is None:
            return _property_not_found("value", json)
        elif not isinstance(value, str):
            return _invalid_property_type("value", "str", json)

        value_type = json.get("valueType")
        if value_type is None:
            return _property_not_found("valueType", json)
        else:
            value_type = DataType.from_str(value_type)
            if isinstance(value_type, DtlError):
                return value_type

        return NodeDescription(path, value, value_type)


class Add(Transformation):
    """
    Adds a node to the graph

    Some edge cases:
        - If the specified path's parent do not exists, the node will not be created
        - If the cast cannot happen for the given value, the node will node be created
    """
    type_str = "Add"

    def __init__(self, nodes: List[NodeDescription]):
        """
        :param nodes: List of nodes to be added to the tree
        """
        Transformation.__init__(self, Add.type_str)
        self.nodes = nodes

    def __eq__(self, other: 'Add'):
        if isinstance(self, other.__class__):
            return self._as_payload() == other._as_payload()
        return False

    def __repr__(self):
        return f"Add(nodes: {self.nodes})"

    def _as_payload(self) -> dict:
        base = self._base_payload()
        base["nodes"] = list(map(lambda x: x._as_payload(), self.nodes))
        return base

    @staticmethod
    def _from_payload(json: dict) -> Union[DtlError, 'Add']:
        nodes = json.get("nodes")
        if nodes is None:
            return _property_not_found("nodes", json)

        nodes = _parse_list(NodeDescription._from_payload)(nodes)
        if isinstance(nodes, DtlError):
            return nodes

        return Add(nodes)



from typing import List, Union, Optional
from datalogue.clients.http import _HttpClient, HttpMethod
from datalogue.models.datastore import Datastore
from datalogue.models.credentials import Credentials, _credentials_ref_from_payload, CredentialsReference
from datalogue.utils import _parse_list
from uuid import UUID
from datalogue.errors import DtlError


class _CredentialsClient:
    """
    Client to interact with the Credentials objects
    """
    def __init__(self, http_client: _HttpClient):
        self.http_client = http_client
        self.service_uri = "/scout/public"
        self.admin = _AdminCredentialsClient(http_client)

    def list(self, credential_type: Optional[str] = None, page: int = 1, item_per_page: int = 25) -> Union[
        DtlError, List[Datastore]]:
        """
        Lists available credentials that user has Use Permission

        :param credential_type: type of the credentials (S3, GCS, Jdbc, etc)
        :param page: page to be retrieved
        :param item_per_page: number of items to be put in a page
        :return: Returns a List of all the available credentials or an error message as a string
        """

        res = self.http_client.make_authed_request(
            self.service_uri +
            f"/credentials?page={page}&size={item_per_page}{'' if credential_type is None else f'&type={credential_type}'}",
            HttpMethod.GET)

        if isinstance(res, DtlError):
            return res

        return _parse_list(_credentials_ref_from_payload)(res)


class _AdminCredentialsClient:
    """
    Client to interact with the Credentials objects
    """

    def __init__(self, http_client: _HttpClient):
        self.http_client = http_client
        self.service_uri = "/scout/admin"

    def create(self, credentials_definition: Credentials, name: Optional[str] = None) -> Union[DtlError, CredentialsReference]:
        """
        Creates the Credentials as specified.

        :param credentials_definition: Credentials Definition to be creates
        :param name: Name to be given to the Credentials to be able to find it later
        :return: string with error message if failed, uuid otherwise
        """
        res = self.http_client.make_authed_request(self.service_uri + "/credentials", HttpMethod.POST, {
            "name": name,
            "credentials": credentials_definition._as_payload()
        })

        if isinstance(res, DtlError):
            return res

        return _credentials_ref_from_payload(res)

    def update(self, credentials_id: UUID, credentials_definition: Credentials,
               name: Optional[str] = None) -> Union[DtlError, CredentialsReference]:
        """
        Updates the given Credentials id with name and credentials

        TODO rework the api so that you can only update the name without the Credentials themselves

        :param credentials_id: id of te Credentials to update
        :param name: if specified, will overwrite current name
        :param credentials_definition: new Credentials definition
        :return:
        """

        body = {}
        if name is not None:
            body["name"] = name

        body["credentials"] = credentials_definition._as_payload()

        res = self.http_client.make_authed_request(
            self.service_uri + "/credentials/" + str(credentials_id),
            HttpMethod.PUT,
            body)

        if isinstance(res, DtlError):
            return res

        return _credentials_ref_from_payload(res)

    def list(self, credential_type: Optional[str] = None, page: int = 1, item_per_page: int = 25) \
            -> Union[DtlError, List[CredentialsReference]]:
        """
        Lists available credentials that Organization has

        :param credential_type: type of the credentials (S3, GCS, Jdbc, etc)
        :param page: page to be retrieved
        :param item_per_page: number of items to be put in a page
        :return: Returns a List of all the available credentials or an error message as a string
        """

        res = self.http_client.make_authed_request(
            self.service_uri +
            f"/credentials?page={page}&size={item_per_page}{'' if credential_type is None else f'&type={credential_type}'}", HttpMethod.GET)

        if isinstance(res, DtlError):
            return res

        return _parse_list(_credentials_ref_from_payload)(res)

    def delete(self, credentials_id: UUID) -> Union[DtlError, bool]:
        """
        Deletes the given Credentials

        :param credentials_id: id of the credentials to be deleted
        :return: true if successful, false otherwise
        """

        res = self.http_client.make_authed_request(
            self.service_uri + "/credentials/" + str(credentials_id),
            HttpMethod.DELETE
        )

        if isinstance(res, DtlError):
            return res
        else:
            return True

    def share(self, credentials_id: UUID, group_id: UUID) -> Union[DtlError, bool]:
        res = self.http_client.make_authed_request(
            self.service_uri + f"/credentials/{str(credentials_id)}/permissions/groups/{str(group_id)}",
            HttpMethod.PUT)

        if isinstance(res, DtlError):
            return res

        return True



# check the input here: https://github.com/datalogue/scout/blob/master/src/main/scala/io/datalogue/scout/api/inputs.scala


from typing import Dict, List, Optional, Union
from datalogue.errors import DtlError
from datalogue.utils import _parse_list
from uuid import UUID


class DatastoreCollection:

    def __init__(self,
                 name: str,
                 storeIds: List[str],
                 description: str = None,
                 tags: List[str] = [],
                 datastore_collection_id: Optional[UUID] = None,
                 owner: Optional[Dict] = None,
                 organization: Optional[Dict] = None,
                 categories: Optional[List] = None,
                 access: Optional[Dict] = None,
                 metadata: Optional[Dict] = None,
                 ):
        """
        Builds pointer to a datastore collection.

        :param name: name of the pointer (mandatory)
        :param DatastoreIds: the datastore ids as a list (mandatory
        :param description: description of the datastore_collection (mandatory)
        :param tags: tags for the datastore_collection (mandatory)
        :param datastore_collection_id: exists if the object was defined
        :param owner:
        :param organization:
        :param categories:
        :param access:
        :param metadata:


        """
        self.name = name
        self.storeIds = storeIds
        self.id = datastore_collection_id
        self.description = description
        self.owner = owner
        self.organization = organization
        self.categories = categories
        self.tags = tags
        self.access = access
        self.metadata = metadata
        self._as_payload()

    def __eq__(self, other: 'DatastoreCollection'):
        if isinstance(self, other.__class__):
            return self._as_payload() == other._as_payload()
        return False

    def _as_payload(self) -> dict:
        """
        Dictionary representation of the object
        :return:
        """
        base = {
            "name": self.name,
            "storeIds": self.storeIds,
            "description": self.description,
            "tags": self.tags
        }

        # toDo unpack proper these values
        if self.id is not None:
            base['id'] = str(self.id)

        if self.owner is not None:
            base["owner"] = self.owner

        if self.organization is not None:
            base["organization"] = self.organization

        if self.categories is not None:
            base["categories"] = self.categories

        if self.access is not None:
            base["access"] = self.access

        if self.metadata is not None:
            base["metadata"] = self.metadata

        return base


def _datastore_collection_from_dict(json: dict) -> Union[DtlError, DatastoreCollection]:
    """Datastore Collection instance from dict."""
    name = json.get("name")
    if name is None:
        return DtlError("'name' for a datastore collection should be defined")

    datastore_collection_id = json.get("id")
    if datastore_collection_id is None:
        return DtlError("'id' for a datastore collection should be defined'")
    datastore_collection_id = UUID(datastore_collection_id)

    tags = json.get("tags")
    if tags is None:
        return DtlError("'tags' for a datastore collection should be defined'")

    #TODO? the `create_datastore_collection` does not return the storeIds

    # Todo?
    # the create function return key "stores"
    if 'stores' in json.keys():
        storeIds = json.get("stores")
    else:
        storeIds = json.get("storeIds")

    description = json.get('description')
    owner = json.get('owner')
    organization = json.get('organization')
    categories = json.get('categories')
    access = json.get('access')
    metadata = json.get('metadata')

    datastore_collection = DatastoreCollection(
                       name,
                       storeIds,
                       description,
                       tags,
                       datastore_collection_id,
                       owner,
                       organization,
                       categories,
                       access,
                       metadata
                    )

    return datastore_collection



from typing import List, Union
from datalogue.errors import DtlError
from datalogue.models.transformations.commons import Transformation

class SplitLabelAndValue(Transformation):
    """
    Finds all nodes that has a label matching to the given regex and attaches two child nodes. 
    One of the node contains the label as a value and the other contains the same value
    """
    type_str = "SplitLabelAndValue"

    def __init__(self, regex: str, labelKey: str, valueKey: str):
        """
        :param regex: a string input to find which nodes should be split into two
        :param labelKey: label of node for label
        :param valueKey: label of node for value
        """
        Transformation.__init__(self, SplitLabelAndValue.type_str)
        self.regex = regex
        self.labelKey = labelKey
        self.valueKey = valueKey

    def __eq__(self, other: 'SplitLabelAndValue'):
        if isinstance(self, other.__class__):
            return self._as_payload() == other._as_payload()
        return False

    def __repr__(self):
        return f"SplitLabelAndValue(regex: {self.regex}, labelKey: {self.labelKey}, valueKey: {self.valueKey})"

    def _as_payload(self) -> dict:
        base = self._base_payload()
        base["regex"] = self.regex
        base["labelKey"] = self.labelKey
        base["valueKey"] = self.valueKey
        return base

    @staticmethod
    def _from_payload(json: dict) -> Union[DtlError, 'SplitLabelAndValue']:
        regex = json.get("regex")
        if not isinstance(regex, str):
            return DtlError("regex field is not a string in %s transformation" % SplitLabelAndValue.type_str)

        labelKey = json.get("labelKey")
        if not isinstance(labelKey, str):
            return DtlError("labelKey field is not a string in %s transformation" % SplitLabelAndValue.type_str)

        valueKey = json.get("valueKey")
        if not isinstance(valueKey, str):
            return DtlError("valueKey field is not a string in %s transformation" % SplitLabelAndValue.type_str)

        return SplitLabelAndValue(regex, labelKey, valueKey)



from typing import Optional, Union, List
from datalogue.clients.http import _HttpClient, HttpMethod
from datalogue.models.ontology import Ontology
from datalogue.utils import _parse_list
from datalogue.errors import DtlError
from uuid import UUID


class _OntologyClient:
    """
    Client to interact with the ontology
    """

    def __init__(self, http_client: _HttpClient):
        self.http_client = http_client
        self.service_uri = "/yggy"

    def create(self, ontology: Ontology) -> Union[DtlError, Ontology]:
        res = self.http_client.make_authed_request(self.service_uri + '/ontology/import', HttpMethod.POST, body=Ontology.as_payload(ontology))

        if isinstance(res, DtlError):
            return res
            
        return Ontology.from_payload(res)

    def get(self, ontology_id: UUID) -> Union[DtlError, Ontology]:
        """
        Get :class:`Ontology` object given ontology_id
        :return: :class:`Ontology` object
        """
        res = self.http_client.make_authed_request(self.service_uri + f'/ontology/{str(ontology_id)}', HttpMethod.GET)

        if isinstance(res, DtlError):
            return res

        return Ontology.from_payload(res)

    def update(self, ontology: Ontology) -> Union[DtlError, Ontology]:
        res = self.http_client.make_authed_request(self.service_uri + f'/ontology/{str(ontology.ontology_id)}', HttpMethod.PUT, body=Ontology.as_payload(ontology))

        if isinstance(res, DtlError):
            return res

        return Ontology.from_payload(res)

    def delete(self, ontology_id: UUID) -> Union[DtlError, bool]:
        """
        Delete :class:`Ontology` based on the given ontology_id
        :return: True if successful, else returns :class:`DtlError`
        """
        res = self.http_client.make_authed_request(self.service_uri + f'/ontology/{str(ontology_id)}', HttpMethod.DELETE)
        if isinstance(res, DtlError):
            return res
        else:
            return True

    def list(self, page: int = 1, item_per_page: int = 25) -> Union[DtlError, List[Ontology]]:
        """
        List the ontologies

        TODO Pagination

        :param page: page to be retrieved
        :param item_per_page: number of items to be put in a page
        :return: Returns a List of all the available Ontologies or an error message as a string

        """
        res = self.http_client.make_authed_request(self.service_uri + f'/ontologies', HttpMethod.GET)

        if isinstance(res, DtlError):
            return res

        return _parse_list(Ontology.from_payload)(res)



from typing import List, Union, Optional
from datalogue.models.datastore import Datastore, DatastoreDef, _datastore_def_from_payload
from datalogue.models.transformations import Transformation, _transformation_from_payload
from datalogue.models.transformations.commons import DataType
from datalogue.utils import _parse_list
from datalogue.errors import DtlError, _property_not_found


class Definition:
    """
    A Pipeline is a series of transformations that will output its results to a target
    and can contain Child(ren ?) Pipelines. cf diagram below.

        Parent Pipelines -> Transformations -+-> Target
                                             |
                                             +-> Children Pipelines
    """
    def __init__(self, transformations: List[Transformation], pipelines: List['Definition'],
                 target: Union[Datastore, DatastoreDef]):
        if isinstance(target, Datastore):
            target = target.definition

        self.transformations = transformations
        self.pipelines = pipelines
        self.target = target

    def __repr__(self):
        return f"Pipeline(type: {self.transformations!r}, pipelines: {self.pipelines!r}, target: {self.target!r})"

    def __eq__(self, other: 'Definition'):
        if isinstance(self, other.__class__):
            return self._as_payload() == other._as_payload()
        return False

    def _as_payload(self) -> Union[DtlError, dict]:
        if self.target.datastore_id is None:
            return DtlError("Cannot serialize a pipeline with a target that was not saved to the database (id missing)")

        return {
            "transformations": list(map(lambda s: s._as_payload(), self.transformations)),
            "pipelines": list(map(lambda s: s._as_payload(), self.pipelines)),
            "target": self.target._as_payload()
        }

    @staticmethod
    def _from_payload(json: dict) -> Union[DtlError, 'Definition']:
        """
        Builds a pipeline object from a dictionary,

        :param json: dictionary parsed from json
        :return:
        """

        transformations = json.get("transformations")
        if transformations is not None:
            transformations = _parse_list(_transformation_from_payload)(transformations)
            if isinstance(transformations, DtlError):
                return transformations
        else:
            transformations = list()

        pipelines = json.get("pipelines")
        if pipelines is not None:
            pipelines = _parse_list(Definition._from_payload)(pipelines)
            if isinstance(pipelines, DtlError):
                return pipelines
        else:
            pipelines = list()

        target = json.get("target")
        if target is None:
            return DtlError("Cannot have a pipeline without a 'target' property")
        else:
            target = _datastore_def_from_payload(target)
            if isinstance(target, DtlError):
                return target

        return Definition(transformations, pipelines, target)


class EnvVariable:
    """
    Object that represents an environment variable.

    An environment variable can be either a literal or an expression that returns a value.

    Right now you can input as a value:
        - a string: "a"
        - an int: 3
        - a float: 4.3
        - a boolean: false
        - a script to index a source of data: `Source(<source_json_object>).index(n => n.label == "some_label")`. This
           will cache the data into a `Map` with the key being the value of the label you specified and value being
           the whole adg.

    The environment variables are injected as arguments into the lambda functions defined inside transformations.

        EX: MapFunction transformation

    ALPHA FEATURE
    """
    def __init__(self, data_type: DataType, value: str, key: str):
        """
        Builds an environment variable to be evaluated before execution of a stream

        :param data_type: type of the variable
        :param value: string to be evaluated
        :param key: key to be used to retrieve the value inside a lambda function
        """
        self.type = data_type
        self.value = value
        self.key = key

    def __eq__(self, other: 'EnvVariable'):
        if isinstance(self, other.__class__):
            return self._as_payload() == other._as_payload()
        return False

    def __repr__(self):
        return f"EnvVariable(type: {self.type.value}, value: {self.value!r}, key: {self.key!r})"

    def _as_payload(self):
        return {
            "type": self.type.value,
            "value": self.value,
            "key": self.key
        }

    @staticmethod
    def _from_payload(json: dict) -> Union['EnvVariable', DtlError]:

        data_type = json.get("type")
        if data_type is None:
            return _property_not_found("type", json)

        data_type = DataType.from_str(data_type)
        if isinstance(data_type, DtlError):
            return data_type

        value = json.get("value")
        if value is None:
            return _property_not_found("value", json)

        key = json.get("key")
        if key is None:
            return _property_not_found("key", json)

        return EnvVariable(data_type, value, key)


class Stream:
    """
    Describes a stream of data flowing from a source to one or several destinations with transformations

        Source -+-> Pipeline 0
                |
                +-> Pipeline 1

               ...
    """
    def __init__(self, source: Union[DatastoreDef, Datastore], pipelines: List[Definition], env: Optional[List[EnvVariable]] = None):
        """
        Creates a stream from a source flowing into different pipelines

        :param source: source of the streaming data
        :param pipelines: pipelines containing the transformations and targets for the data
        :param env: list of environment variables to be evaluated before beginning streaming the data
        """
        if isinstance(source, Datastore):
            source = source.definition

        self.source = source
        self.pipelines = pipelines
        self.env = env

    def __repr__(self):
        return f"Stream(type: {self.source!r}, pipelines: {self.pipelines!r}, env: {self.env!r})"

    def __eq__(self, other: 'Stream'):
        if isinstance(self, other.__class__):
            return self._as_payload() == other._as_payload()
        return False

    def _as_payload(self):
        base =  {
            "source": self.source._as_payload(),
            "pipelines": list(map(lambda s: s._as_payload(), self.pipelines))
        }
        if self.env is not None:
            base["env"] = list(map(lambda s: s._as_payload(), self.env))

        return base

    @staticmethod
    def _from_payload(json: dict) -> Union[DtlError, 'Stream']:
        """
        Builds a Stream instance from a json object

        :param json:
        :return: if fails returns a string with the error message
        """
        source = json.get("source")
        if source is None:
            return DtlError("stream needs a source of data")
        else:
            source = _datastore_def_from_payload(source)
            if isinstance(source, DtlError):
                return source

        pipelines = json.get("pipelines")
        if pipelines is None:
            return DtlError("streams needs a 'pipelines' property")
        else:
            pipelines = _parse_list(Definition._from_payload)(pipelines)
            if isinstance(pipelines, DtlError):
                return pipelines

        env = json.get("env")
        if env is not None:
            env = _parse_list(EnvVariable._from_payload)(env)
            if isinstance(env, DtlError):
                return env

        return Stream(source, pipelines, env)



from abc import ABC, abstractmethod
from typing import Dict, List, Set, Optional, Union
from uuid import UUID
import itertools

from datalogue.errors import _enum_parse_error, DtlError
from datalogue.utils import _parse_list, _parse_string_list, SerializableStringEnum


class OntologyNode:
    def __init__(self,
                 name: str,
                 description: Optional[str] = None,
                 children: List['OntologyNode'] = [],
                 node_id: Optional[UUID] = None):

        if not isinstance(name, str):
            raise DtlError("name should be string in OntologyNode")

        if description is not None and not isinstance(description, str):
            raise DtlError("description should be string in OntologyNode")

        if not isinstance(children, List):
            raise DtlError("children should be list in OntologyNode")

        if node_id is not None and not isinstance(node_id, UUID):
            raise DtlError("node_id should be uuid in OntologyNode")

        self.name = name
        self.description = description
        self.children = children
        self.node_id = node_id

    def __eq__(self, other: 'OntologyNode'):
        if isinstance(self, other.__class__):
            return (self.name == other.name and
                    self.description == other.description and
                    self.children == other.children and
                    self.node_id == other.node_id)
        return False

    @staticmethod
    def as_payload(ontology_node: 'OntologyNode') -> Union[DtlError, dict]:
        payload = {
            "name": ontology_node.name,
            "children": list(map(lambda n: OntologyNode.as_payload(n), ontology_node.children))
        }

        if ontology_node.description is not None:
            payload["description"] = ontology_node.description

        return payload

    @staticmethod
    def from_payload(payload: dict) -> Union[DtlError, 'OntologyNode']:
        node_id = UUID(payload.get("id"))
        name = payload.get("name")
        description = payload.get("description")
        children = payload.get("children")
        if children is None:
            children = []
        else:
            children = _parse_list(OntologyNode.from_payload)(children)
        return OntologyNode(name, description, children, node_id=node_id)


class Ontology:
    def __init__(self,
                 name: str,
                 description: Optional[str] = None,
                 tree: List[OntologyNode] = [],
                 ontology_id: Optional[UUID] = None):

        if not isinstance(name, str):
            raise DtlError("name should be string in Ontology")

        if description is not None and not isinstance(description, str):
            raise DtlError("description should be string in Ontology")

        if not isinstance(tree, List):
            raise DtlError("tree should be list in Ontology")

        if ontology_id is not None and not isinstance(ontology_id, UUID):
            raise DtlError("ontology_id should be uuid in Ontology")

        self.name = name
        self.description = description
        self.tree = tree
        self.ontology_id = ontology_id



    def __eq__(self, other: 'Ontology'):
        if isinstance(self, other.__class__):
            return (self.name == other.name and
                    self.description == other.description and
                    self.tree == other.tree and
                    self.ontology_id == other.ontology_id)
        return False

    def __repr__(self):
        def print_nodes(tree, output=[], level=0):
            for n in tree:
                padding = '   ' * level
                if level == 0:
                    output.append(padding + n.name)
                else:
                    output.append(padding + '|___' + n.name)
                for c in n.children:
                    print_nodes([c], output, level=level+1)
            return output

        first_line = f'Ontology(id: {self.ontology_id}, name: {self.name!r}, description: {self.description!r})' + '\n'
        return '\n'.join(print_nodes(self.tree, [first_line]))

    def leaves(self) -> List[OntologyNode]:
        def iterate(node: OntologyNode) -> List[OntologyNode]:
            if not node.children:
                return [ node ]
            else:
                return list(itertools.chain(*map(lambda n: iterate(n), node.children)))

        return list(itertools.chain(*map(lambda n: iterate(n), self.tree)))

    @staticmethod
    def as_payload(ontology: 'Ontology') -> Union[DtlError, dict]:
        payload = {
            "name": ontology.name,
            "children": list(map(lambda n: OntologyNode.as_payload(n), ontology.tree))
        }

        if ontology.description is not None:
            payload["description"] = ontology.description

        return payload

    @staticmethod
    def from_payload(payload: dict) -> Union[DtlError, 'Ontology']:
        ontology_id = UUID(payload.get("id"))
        name = payload.get("name")
        description = payload.get("description")
        tree = payload.get("tree")
        if tree is None:
            tree = []
        else:
            tree = _parse_list(OntologyNode.from_payload)(payload["tree"])
        return Ontology(name, description, tree, ontology_id=ontology_id)


class DataRef:
    def __init__(self, node: OntologyNode, path_list: List[List[str]]):
        self.node = node
        self.path_list = path_list



from typing import List, Union
from uuid import UUID
from datalogue.models.stream import Stream
from datalogue.utils import _parse_list
from datalogue.errors import DtlError


class StreamMetadata:
    def __init__(self, stream_id: UUID, is_ready: bool, stream: Stream):
        self.id = stream_id
        self.is_ready = is_ready
        self.stream = stream

    def __repr__(self):
        return f'{self.__class__.__name__}(id: {self.id},' \
               f' is_ready: {self.is_ready!r}, stream: {self.stream!r})'

    def _as_payload(self):
        base = dict()
        base["id"] = str(self.id)
        base["isReady"] = self.is_ready
        base["stream"] = self.stream._as_payload()
        return base


def _stream_metadata_from_payload(json: dict) -> Union[DtlError, StreamMetadata]:
    stream_id = json.get("id")
    if stream_id is None:
        return DtlError("'id' not defined in stream metadata")
    else:
        try:
            stream_id = UUID(stream_id)
        except ValueError:
            return DtlError("'id' is not a valid uuid")

    is_ready = json.get("isReady")
    if is_ready is None:
        is_ready = False

    stream = json.get("stream")
    if stream is None:
        return DtlError("'stream' needs to be defined in stream metadata payload")
    else:
        stream = Stream._from_payload(stream)
        if isinstance(stream, DtlError):
            return stream

    return StreamMetadata(stream_id, is_ready, stream)


class StreamCollection:
    def __init__(self, stream_collection_id: UUID, name: str, streams: List[StreamMetadata]):
        self.id = stream_collection_id
        self.name = name
        self.streams = streams

    def __repr__(self):
        return f'{self.__class__.__name__}(id: {self.id}, name: {self.name!r}, streams: {self.streams!r})'

    def _as_payload(self) -> dict:
        """
        Dictionary representation of the object with camelCase keys
        :return:
        """
        base = dict()
        base["id"] = str(self.id)
        base["name"] = self.name
        base["streams"] = list(map(lambda s: s._as_payload(), self.streams))
        return base


def _stream_collection_from_payload(json: dict) -> Union[DtlError, StreamCollection]:
    """
    Builds a Stream Collection from a json dictionary

    :param json:
    :return: a string with the error message or the Stream Collection
    """

    stream_collection_id = json.get("id")
    if stream_collection_id is None:
        return DtlError("A Stream Collection needs to have an 'id' property")
    else:
        try:
            stream_collection_id = UUID(stream_collection_id)
        except ValueError:
            return DtlError("The id field in the Stream Collection dictionary was not a valid UUID")

    name = json.get("name")
    if name is None:
        return DtlError("A Stream Collection needs a 'name' field")
    else:
        if not isinstance(name, str):
            return DtlError("name field in the Stream Collection is expected to be a string")

    streams = json.get("streams")
    if streams is None:
        return DtlError("A Stream Collection needs a 'streams' property")
    else:
        streams = _parse_list(_stream_metadata_from_payload)(json["streams"])
        if isinstance(streams, DtlError):
            return streams

    return StreamCollection(stream_collection_id, name, streams)



from typing import List, Union
from uuid import UUID

from datalogue.clients.http import _HttpClient, HttpMethod
from datalogue.models.organization import Organization, _organization_from_payload, User, _users_from_payload, Group, _group_from_payload, Domain, _domain_from_payload
from datalogue.errors import DtlError
from datalogue.utils import _parse_list


class _OrganizationClient:
    """
    Client to interact with the Organization objects
    """
    errorMsg = "you may not have permission to do the action or the organization was not found"

    def __init__(self, http_client: _HttpClient):
        self.http_client = http_client

    def create(self, name: str) -> Union[DtlError, Organization]:
        """
        Creates an organization

        :param name: Organization to be created
        :return: string with error message if failed, the organization otherwise
        """
        res = self.http_client.make_authed_request(
            "/organization", HttpMethod.POST, {
                "name": name
            })

        if isinstance(res, DtlError):
            return DtlError("Could not create the organization:", res)

        return _organization_from_payload(res)

    def get(self, org_id: UUID) -> Union[DtlError, Organization]:
        """
        From the provided id, get the corresponding Organization

        :param org_id: id of the organization to be retrieved
        :return: Error if it fails or Organization object otherwise
        """

        res = self.http_client.make_authed_request(
            f"/organization/{str(org_id)}/public", HttpMethod.GET)

        if isinstance(res, DtlError):
            return DtlError("Could not retreive the organization:", res)

        return _organization_from_payload(res)

    def get_all_users(self, org_id: UUID) -> Union[DtlError, List[User]]:
        """
        From the provided organization id, get all users in the Organization

        :param org_id: id of the organization
        :return: Error if it fails or list of all users in the organization otherwise
        """

        res = self.http_client.make_authed_request(
            f"/organization/{str(org_id)}/users", HttpMethod.GET)

        if isinstance(res, DtlError):
            return DtlError("Could not retreive all users of the organization:", _OrganizationClient.errorMsg)

        return _parse_list(_users_from_payload)(res)

    def get_groups(self, org_id: UUID) -> Union[DtlError, List[Group]]:
        """
        From the provided organization id, get all groups in the Organization in which the user is a member

        :param org_id: id of the organization
        :return: Error if it fails or list of all groups in the organization otherwise
        """

        res = self.http_client.make_authed_request(
            f"/groups?orgId={str(org_id)}", HttpMethod.GET)

        if isinstance(res, DtlError):
            return DtlError("Could not retreive your groups of the organization:")

        return _parse_list(_group_from_payload)(res)

    def get_all_groups(self, org_id: UUID) -> Union[DtlError, List[Group]]:
        """
        From the provided organization id, get all groups in the Organization (should be admin)

        :param org_id: id of the organization
        :return: Error if it fails or list of all groups in the organization otherwise
        """

        res = self.http_client.make_authed_request(
            f"/organization/{str(org_id)}/groups", HttpMethod.GET)

        if isinstance(res, DtlError):
            return DtlError("Could not retrieve all groups of the organization:", _OrganizationClient.errorMsg)

        return _parse_list(_group_from_payload)(res)

    def add_domain(self, org_id: UUID, domain: str) -> Union[DtlError, bool]:
        """
        Add domain to provided organization

        :param org_id: id of the organization
        :return: Error if it fails or true if it's added
        """

        res = self.http_client.make_authed_request(
            f"/organization/{str(org_id)}/domain", HttpMethod.POST, {
                "domain": domain
            })

        if isinstance(res, DtlError):
            return DtlError("Could not add domain to the organization:", _OrganizationClient.errorMsg)
        else:
            return True

    def remove_domain(self, org_id: UUID, domain: str) -> Union[DtlError, bool]:
        """
        Remove domain from provided organization

        :param org_id: id of the organization
        :return: Error if it fails or true if it's deleted
        """

        res = self.http_client.make_authed_request(
            f"/organization/{str(org_id)}/domain", HttpMethod.DELETE, {
                "domain": domain
            })

        if isinstance(res, DtlError):
            return DtlError("Could not delete domain:", _OrganizationClient.errorMsg)
        else:
            return True

    def get_all_domains(self, org_id: UUID) -> Union[DtlError, List[Domain]]:
        """
        Get all domains of the provided organization

        :param org_id: id of the organization
        :return: Error if it fails or list of domains
        """

        res = self.http_client.make_authed_request(
            f"/organization/{str(org_id)}/domains", HttpMethod.GET)

        if isinstance(res, DtlError):
            return DtlError("Could not retreive all domains:", _OrganizationClient.errorMsg)

        return _parse_list(_domain_from_payload)(res)




from typing import Optional
from uuid import UUID, uuid4
from pathlib import Path

from datalogue.models.metrics import ArgoMetrics
from datalogue.clients.http import _HttpClient, HttpMethod
from datalogue.errors import DtlError


class _MetricsClient:
    """
    Client to interact with the Metrics View of a Training

    This is meant to be used by Argo Science Mode users
    as an interface to quickly display their results to customers.
    """

    def __init__(self, http_client: _HttpClient):
        self.http_client = http_client

    def upload(self, ontology_id: UUID, name: str, argo_work_dir: Path, experiment_name: str, class_map_path: str, training_id: Optional[UUID] = None) -> Optional[bool]:
        """
        Processes and pushes via the Argus api to the defined
        ontology.

        :param ontology_id: UUID the ontology you are publishing to
        :param name: str who is pushing this data
        :param argo_work_dir: str Path to the template's work directory (i.e. nightshade/work/cbc)
        :param experiment_name: str Name of the experiment to upload
        :param class_map_path: str
        :return: Error if it fails or True otherwise
        """
        if training_id is None:
            training_id = str(uuid4())

        M = ArgoMetrics(str(argo_work_dir), experiment_name, class_map_path)
        endpoint = f'/yggy/ontology/{ontology_id}/trainings/upload-metrics'
        params = {
            "training-id": f"{training_id}",
            "user-firstname": f"{name}",
            "user-lastname": "SDK",
        }

        res = self.http_client.make_authed_request(endpoint, HttpMethod.PUT, M.payload, params)

        if isinstance(res, DtlError):
            return DtlError("Could not publish request.", res)

        return True



from typing import List, Union
from datalogue.errors import DtlError
from datalogue.models.transformations.commons import Transformation, _array_from_dict

class MoveByRegex(Transformation):
    """
    Finds all nodes that has a label matching to the given regex 
    and moves them as children of a new parent identified by the given to path
    """
    type_str = "MoveByRegex"

    def __init__(self, regex: str, to: List[str]):
        """
        :param regex: a string input to find which nodes should be split into two
        :param to: array of string
        """
        Transformation.__init__(self, MoveByRegex.type_str)
        self.to = to
        self.regex = regex

    def __eq__(self, other: 'MoveByRegex'):
        if isinstance(self, other.__class__):
            return self._as_payload() == other._as_payload()
        return False

    def __repr__(self):
        return f"MoveByRegex(regex: {self.regex}, to: {'.'.join(self.to)})"

    def _as_payload(self) -> dict:
        base = self._base_payload()
        base["regex"] = self.regex
        base["to"] = self.to
        return base

    @staticmethod
    def _from_payload(json: dict) -> Union[DtlError, 'MoveByRegex']:
        regex = json.get("regex")
        if not isinstance(regex, str):
            return DtlError("regex field is not a string in %s transformation" % MoveByRegex.type_str)

        to = _array_from_dict(json, MoveByRegex.type_str, "to")
        if isinstance(to, DtlError):
            return to

        return MoveByRegex(regex, to)



from codecs import open
from os import path

from setuptools import find_packages, setup

here = path.abspath(path.dirname(__file__))

# Get the long description from the README file 
with open(path.join(here, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

dtl_version = {}
with open("./datalogue/version.py") as fp:
    exec(fp.read(), dtl_version)

with open("./LICENSE", "r") as license_file:
    imported_license = license_file.read()
    setup(
        name="datalogue",
        version=dtl_version["__version__"],
        author="Nicolas Joseph",
        author_email="nic@datalogue.io",
        license=imported_license,
        description="SDK to interact with the datalogue platform",
        long_description_content_type="text/markdown",
        url="https://github.com/datalogue/dtl-python-sdk",
        packages=find_packages(exclude=["*.tests", "*.tests.*", "tests.*", "tests"]),
        classifiers=[
            "Programming Language :: Python :: 3",
            'Programming Language :: Python :: 3.6',
            "Operating System :: OS Independent"
        ],
        python_requires=">=3.6",
        install_requires=['requests', 'python-dateutil', 'validators', 'pytest==3.6.3', 'numpy', 'pyyaml', 'pyarrow', 'raven'],
        setup_requires=['pytest-runner'],
        tests_require=['pytest==3.6.3', 'pytest-cov==2.6.0']
)



import os
import json
import yaml
import numpy as np
from glob import glob
from typing import List, Union

from datalogue.errors import DtlError


class ArgoMetrics:
    def __init__(self, argo_work_dir: str, experiment_name: str, class_map_path: str):
        if not os.path.isdir(argo_work_dir):
            raise DtlError(
                f"{argo_work_dir} is not a directory. Path to the template's work directory(i.e. nightshade/work/cbc"
            )
        self.argo_work_dir = argo_work_dir
        self.experiment_name = experiment_name
        self.class_map = self._load_class_map(class_map_path)
        self.training_files = self._fetch_training_files()
        self.test_file = self._fetch_test_file()
        self.payload = self.create_payload()

    def _fetch_training_files(self) -> List[str]:
        tfs = sorted(glob(os.path.join(self.argo_work_dir, self.experiment_name, "control",
                                       "*-train-completed", "epoch-*-experiment-tracking.json")))
        if len(tfs) == 0:
            raise DtlError(
                f"No Training Files found in {self.argo_work_dir}, experiment: {self.experiment_name}")
        return tfs

    def _fetch_test_file(self) -> str:
        tfs = sorted(glob(os.path.join(self.argo_work_dir, self.experiment_name, "control",
                                       "*-test-completed", "test-results.json")))
        if len(tfs) == 0:
            raise DtlError(
                f"No Test Files found in {self.argo_work_dir}, experiment: {self.experiment_name}")
        return tfs[-1]

    @staticmethod
    def _load_class_map(class_map_path: str) -> dict:
        if not os.path.isfile(class_map_path):
            raise DtlError(
                f"{class_map_path} is not a file. Please provide a Json or Yaml file!"
            )
        with open(class_map_path, 'r', encoding='utf-8') as f:
            if class_map_path.endswith('json'):
                return json.load(f)
            elif class_map_path.endswith('yaml'):
                return yaml.load(f)
            raise DtlError(
                f"{class_map_path} is not Json or Yaml. Unable to Load Class Map")

    def get_metrics(self, conf_mat: List[List], phase_name: str, class_names: List[str]) -> dict:
        rev_class_map = {v: k for k, v in self.class_map.items()}
        fp = conf_mat.sum(axis=0) - np.diag(conf_mat)
        fn = conf_mat.sum(axis=1) - np.diag(conf_mat)
        tp = np.diag(conf_mat)
        tn = conf_mat.sum() - (fp + fn + tp)
        epsilon = 1e-15  # So that we cannot get division by 0 errors
        # Sensitivity, hit rate, recall, or true positive rate
        tpr = tp / (tp + fn + epsilon)
        fpr = fp / (fp + tn + epsilon)
        recall = {rev_class_map[i]: float(tpr[i]) for i in range(len(tpr))}
        # Precision or positive predictive value
        ppv = tp / (tp + fp + epsilon)
        precision = {rev_class_map[i]: float(ppv[i]) for i in range(len(ppv))}
        # f1 Score
        f1 = 2 * tpr * ppv / (tpr + ppv + epsilon)
        f1_score = {rev_class_map[i]: float(f1[i]) for i in range(len(f1))}

        evaluation = {'f1': f1_score,
                      'recall': recall,
                      'precision': precision,
                      'fpr': np.mean(fpr),
                      'tpr': np.mean(tpr)
                      }

        return evaluation

    def create_payload(self) -> List[dict]:
        payload = []
        for file in self.training_files:
            with open(file, 'r') as f:
                data = json.load(f)
            M = {"epoch": data["epoch"], 'metrics': {}}
            for phase in ["training", "validation"]:
                M['metrics'][phase] = {
                    "loss": data["metrics"][phase]["loss"],
                    "acc": data["metrics"][phase]["acc"],
                    "avg_precision_loss": data["metrics"][phase]["avg_precision_loss"],
                    "precision":  data["metrics"][phase]["precision"],
                    "recall":  data["metrics"][phase]["recall"],
                    "f1": data["metrics"][phase]["f1"],
                    "tpr": data["metrics"][phase]["tpr"] if "tpr" in data["metrics"][phase] else None,
                    "fpr": data["metrics"][phase]["fpr"] if "fpr" in data["metrics"][phase] else None,
                    "confusion_matrix": None if "training" == phase else data["metrics"][phase]["confusion_matrix"]
                }
                if "validation" == phase:
                    conf_mat = np.array(
                        data["metrics"][phase]["confusion_matrix"]["matrix"])
                    evaluation = self.get_metrics(
                        conf_mat, phase_name=f"{self.experiment_name}_VAL_epoch_{data['epoch']}_", class_names=data["metrics"][phase]["confusion_matrix"]["index"])
                    M['metrics'][phase].update(evaluation)
            payload.append(M)

        with open(self.test_file, 'r') as f:
            test_data = json.load(f)

        M = {"epoch": len(self.training_files), 'metrics': {}}
        for phase in ["training", "validation"]:
            M['metrics'][phase] = {
                "loss": test_data["metrics"]["test"]["loss"],
                "acc": test_data["metrics"]["test"]["acc"],
                "avg_precision_loss": test_data["metrics"]["test"]["avg_precision_loss"],
                "precision": test_data["metrics"]["test"]["precision"],
                "recall": test_data["metrics"]["test"]["recall"],
                "f1": test_data["metrics"]["test"]["f1"],
                "confusion_matrix": test_data["metrics"]["test"]["confusion_matrix"]
            }

            conf_mat = np.array(test_data["metrics"]["test"]
                                ["confusion_matrix"]["matrix"])
            evaluation = self.get_metrics(conf_mat,
                                          phase_name=f"{self.experiment_name}_test_",
                                          class_names=test_data["metrics"]["test"]["confusion_matrix"]["index"])
            M['metrics'][phase].update(evaluation)
        payload.append(M)
        return payload



from typing import List, Union, Dict, Optional
from datalogue.errors import DtlError, _property_not_found, _enum_parse_error, _invalid_property_type
from datalogue.models.transformations.commons import Transformation, DataType
from datalogue.utils import _parse_string_list, SerializableStringEnum, _parse_list


class PickStrategy(SerializableStringEnum):
    """
    Data Types that can be specified in the pipeline
    """
    HighScore = "HighScore"
    Non = "None"
    Random = "Random"
    All = "All"

    @staticmethod
    def parse_error(s: str) -> str:
        return _enum_parse_error("PickStrategy", s)

    @staticmethod
    def from_str(string: str) -> Union[DtlError, 'PickStrategy']:
        return SerializableStringEnum.from_str(PickStrategy)(string)


class ClassNodeDescription:
    """
    Description of a node that will be selected from a class and that specifies its path location in the output.
    """
    def __init__(self, path: List[str], tag: str, pick_strategy: PickStrategy, data_type: DataType = DataType.String):
        """
        Builds a ClassNodeDescription to be used in the Structure transformation.

        :param path: path of the node to be created
        :param tag: class of the value to be picked up
        :param pick_strategy: How to pick up the value to be used
        :param data_type: Type of the output value
        """
        self.path = path
        self.tag = tag
        self.pick_strategy = pick_strategy
        self.data_type = data_type

    def __eq__(self, other: 'ClassNodeDescription'):
        if isinstance(self, other.__class__):
            return self._as_payload() == other._as_payload()
        return False

    def __repr__(self):
        return f"ClassNodeDescription(path: {self.path}, tag: {self.tag}, strategy: {self.pick_strategy}, dataType: {self.data_type})"

    def _as_payload(self) -> dict:
        return {
            "type": "Class",
            "outputDataType": self.data_type.value,
            "outputPath": self.path,
            "class": self.tag,
            "pickStrategy": self.pick_strategy.value
        }

    @staticmethod
    def _from_payload(json: dict) -> Union[DtlError, 'ClassNodeDescription']:
        if json.get("type") is None and json.get("type") != "Class":
            return DtlError("Invalid json payload type for ClassNodeDescription")

        path = json.get("outputPath")
        if path is None:
            return _property_not_found("outputPath", json)

        path = _parse_string_list(path)
        if isinstance(path, DtlError):
            return path

        tag = json.get("class")
        if tag is None:
            return _property_not_found("class", json)

        pick = json.get("pickStrategy")
        if pick is None:
            return _property_not_found("pickStrategy", json)

        pick = PickStrategy.from_str(pick)
        if isinstance(pick, DtlError):
            return pick

        data_type = json.get("outputDataType")
        if data_type is not None:
            data_type = DataType.from_str(data_type)
            if isinstance(data_type, DtlError):
                return data_type
        else:
            data_type = DataType.String

        return ClassNodeDescription(path, tag, pick, data_type)


class PathNodeDescription:
    """
    Instruction to copy a node from input into the output
    """
    def __init__(self, path: List[str], output_path: List[str], output_data_type: DataType = DataType.String,
                 is_output_array: bool = False):
        """
        Builds a PathNodeDescription to be used in the Structure transformation.

        :param path: path of the node to be used as input
        :param output_path: path of the node to be created in the output
        :param output_data_type: type for the node to be used as output
        :param is_output_array: will the output contain several nodes with identical paths
        """
        self.path = path
        self.output_path = output_path
        self.output_dataType = output_data_type
        self.is_output_array = is_output_array

    def __eq__(self, other: 'PathNodeDescription'):
        if isinstance(self, other.__class__):
            return self._as_payload() == other._as_payload()
        return False

    def __repr__(self):
        return f"PathNodeDescription(path: {self.path}, output_path: {self.output_path}, output_dataType: {self.output_dataType}, is_output_array: {self.is_output_array})"

    def _as_payload(self) -> dict:
        return {
            "type": "Path",
            "path": self.path,
            "outputPath": self.output_path,
            "outputDataType": self.output_dataType.value,
            "outputIsArray": self.is_output_array
        }

    @staticmethod
    def _from_payload(json: dict) -> Union[DtlError, 'PathNodeDescription']:
        if json.get("type") is None and json.get("type") != "Path":
            return DtlError("Invalid json payload type for PathNodeDescription")

        path = json.get("path")
        if path is None:
            return _property_not_found("path", json)

        path = _parse_string_list(path)
        if isinstance(path, DtlError):
            return path

        output_path = json.get("outputPath")
        if output_path is None:
            return _property_not_found("outputPath", json)

        output_path = _parse_string_list(output_path)
        if isinstance(output_path, DtlError):
            return output_path

        is_output_array = json.get("is_output_array")
        if is_output_array is None:
            is_output_array = False

        output_data_type = json.get("outputDataType")
        if output_data_type is None:
            return _property_not_found("output_dataType", json)

        output_data_type = DataType.from_str(output_data_type)
        if isinstance(output_data_type, DtlError):
            return output_data_type

        return PathNodeDescription(path, output_path, output_data_type, is_output_array)


class Structure(Transformation):
    """
    Enables to pick nodes based on their class or path and organize them as you wish for the output

    When using a class to select a node, there are different strategies that can be used to pick it.

       * HighScore: Uses the node classified with the highest score
       * None: If several matches do not use any
       * Random: Just picks one
       * All: Picks all

    If no nodes with the specified class or path are found, the output node will be created with an empty value.

    If a cast fails, the element is dropped from the stream.
    """

    type_str = "Structure"

    def __init__(self, structure: List[Union[ClassNodeDescription, PathNodeDescription]]):
        """
        Builds StructureByClass transformation

        :param structure: List of nodes to include in the output
        """
        Transformation.__init__(self, Structure.type_str)
        self.structure = structure

    def __eq__(self, other: 'Structure'):
        if isinstance(self, other.__class__):
            return self._as_payload() == other._as_payload()
        return False

    def __repr__(self):
        return f"Structure(structure: {self.structure!r})"

    def _as_payload(self) -> dict:
        base = self._base_payload()
        base["outputSchema"] = list(map(lambda s: s._as_payload(), self.structure))
        return base

    @staticmethod
    def _parse_class_or_path_description(json: Dict) -> Union[ClassNodeDescription, PathNodeDescription, DtlError]:
        description_type = json.get("type")
        if json.get("type") is None:
            return DtlError("Invalid node description inside the Structure transformation")

        if description_type == "Class":
            return ClassNodeDescription._from_payload(json)
        elif description_type == "Path":
            return PathNodeDescription._from_payload(json)
        else:
            return DtlError("Invalid type for the node description")

    @staticmethod
    def _from_payload(json: dict) -> Union[DtlError, 'Structure']:
        structure = json.get("outputSchema")
        if structure is None:
            return _property_not_found("outputSchema", json)

        structure = _parse_list(Structure._parse_class_or_path_description)(structure)
        if isinstance(structure, DtlError):
            return structure

        return Structure(structure)



from typing import TypeVar, List, Callable, Union, Dict, Type
from enum import Enum
from abc import abstractmethod, ABC
from io import BytesIO, SEEK_SET, SEEK_END

from datalogue.errors import DtlError

T = TypeVar('T')
Json = Union[Dict, str, int, List]


def _parse_list(parse_function: Callable[[Json], Union[DtlError, T]]) -> Callable[[List[Json]], Union[DtlError, List[T]]]:
    """
    Returns another function that can be used to parse a list of json with the specified parse function

    :param parse_function: function to be used for parsing
    :return:
    """
    def parse_concrete(objects: List[Json]) -> Union[DtlError, List[T]]:
        """
        Applies the parse function specified in the parent function to each element of the list of json

        :param objects: list to apply the parse function to
        :return:
        """
        parsed_list = []
        for obj in objects:
            parsed_obj = parse_function(obj)

            if isinstance(parsed_obj, DtlError):
                return parsed_obj
            else:
                parsed_list.append(parsed_obj)

        return parsed_list
    return parse_concrete


def _parse_string_list(objects: List[Json]) -> Union[DtlError, List[str]]:
    parsed_list = []

    for obj in objects:
        if isinstance(obj, str):
            parsed_list.append(obj)
        else:
            return DtlError("The following object is not a string: %s" % obj)

    return parsed_list


def not_implemented() -> NotImplemented:
    return NotImplemented


class SerializableStringEnum(Enum):

    @staticmethod
    def from_str(enum: Type['SerializableStringEnum']) -> Callable[[str], Union[DtlError, 'SerializableStringEnum']]:
        """
        Builds a function to parse the string enum `enum`
        :param enum: enum to create a parser for
        :return:
        """
        def inner_sanctum(s: str) -> Union[DtlError, 'SerializableStringEnum']:
            """
            Parses a string and returns the instance of the Enum it corresponds to or a string with an error message

            :param s: string to be parsed
            :return:
            """
            for blob_type in enum:
                if blob_type.value == s:
                    return blob_type

            return enum.parse_error(s)

        return inner_sanctum

    def __repr__(self):
        return f"{self._value_}"

    @staticmethod
    @abstractmethod
    def parse_error(s: str) -> DtlError:
        """
        Returns the error to be used if the parsing fails
        :return: string error
        """
        return NotImplemented


class ResponseStream(object):

    def __init__(self, request_iterator):
        """
        Class used internally to be able to transform a `requests` iterator into a File like object.

        :param request_iterator: Iterator to pull the data from
        """

        # In memory bytes buffer
        self._bytes = BytesIO()
        # Iterator to use to retrieve data
        self._iterator = request_iterator

        self.closed = self._bytes.closed

    def close(self):
        return self._bytes.close()

    def _load_all(self) -> int:
        """
        Loads all the data into the buffer

        :return: new absolute position in the buffer
        """
        self._bytes.seek(0, SEEK_END)
        for chunk in self._iterator:
            self._bytes.write(chunk)

    def _load_until(self, goal_position: int) -> None:
        """
        Loads the data into the inner buffer until the goal position

        :param goal_position: Byte index to go to
        :return:
        """

        # End of the byte buffer
        current_position = self._bytes.seek(0, SEEK_END)

        # Fill the buffer with more data until we reach the new byte index goal
        while current_position < goal_position:
            try:
                current_position += self._bytes.write(next(self._iterator))
            except StopIteration:
                break

    def tell(self) -> int:
        """
        Current file position, an integer.

        :return:
        """
        return self._bytes.tell()

    def read(self, size=None) -> bytes:
        """
        Read at most size bytes, returned as a bytes object.

        :param size: in bytes to retrieve, if negative reads all
        :return: byte object
        """
        left_off_at = self._bytes.tell()

        # Fill the buffer with new data from the iterator
        if size is None:
            self._load_all()
        else:
            goal_position = left_off_at + size
            self._load_until(goal_position)

        # make sure we are reading from the right position
        self._bytes.seek(left_off_at)

        # return the size bytes from the buffer
        return self._bytes.read(size)

    def seek(self, position, whence=SEEK_SET) -> int:
        """

        :param position:
        :param whence:
        :return:
        """
        if whence == SEEK_END:
            self._load_all()
            return self._bytes.seek(position, whence)
        else:
            return self._bytes.seek(position, whence)
